package sdsi;

import sdsi.sexp.*;

public class Auth extends Cert {
  /** Creates a new Auth from an Issuer, a Subject, a Tag,  
      a propagate boolean, and a Validity*/
  public Auth(SDSIPrincipal p, Subject s, Tag t, boolean prop, Validity v) {
    super(p,s,t,prop,v);
    int pExists = prop ? 1 : 0;
    SexpList propPart = new SexpList(new Sexp[] {new SexpString("propagate")});
    Sexp [] issuerParts = new Sexp[2];
    issuerParts[0] = new SexpString("issuer");
    issuerParts[1] = p.getSrep();
    Sexp[] subjParts = new Sexp[2];
    subjParts[0] = new SexpString("subject");
    subjParts[1] = s.getSrep();
    Sexp[] valids = v.getSreps();
    Sexp[] authParts = new Sexp[4 + pExists + valids.length];
    authParts[0] = new SexpString("cert");
    authParts[1] = new SexpList(issuerParts);
    authParts[2] = new SexpList(subjParts);
    if (prop)
      authParts[3] = propPart;
    authParts[3+pExists] = t.getSrep();
    for (int i=0;i<valids.length;i++)
      authParts[4+pExists+i] = valids[i];
    srep = new SexpList(authParts);
  }
  public Auth(SexpList l, SDSIPrincipal issuer) throws SexpParseException {
    super(l);
    boolean delegateDefined, issuerLocDefined, subjectLocDefined;
    this.issuer = issuer;
    issuerLocDefined = subjectLocDefined = delegateDefined = false;
    Sexp[] certParts = l.arrayOfElements(); 
    for (int j = 1; j < certParts.length; j++) {
      Sexp element = certParts[j];
      if (element instanceof SexpList) {
        SexpList list = (SexpList) element;
        String label = list.getType();
        if (label.equals("tag")) {
      	  if (t != null)
      	    throw new SexpParseException("More than one tag in cert");
      	  else
      	    t = new Tag(list);
      	}
      	else if (label.equals("subject-info")) {
      	  if (subjectLocDefined)
      	    throw new SexpParseException("More than one Subject info in cert");
      	  else {
      	    subjectLocDefined = true;
      	    if (!list.allElementsStrings())
      	      throw new SexpParseException("Subject info should contain only uri's");
      	  }
      	}
      	else if (label.equals("issuer-info")) {
      	  if (issuerLocDefined)
                   throw new SexpParseException("More than one Issuer info in cert");
      	  else {
      	    issuerLocDefined = true;
                  if (!list.allElementsStrings())
      	      throw new SexpParseException("Issuer info should contain only uri's");
      	  }
      	}   
      	else if (label.equals("propagate")) {
      	  if (delegateDefined)
      	    throw new SexpParseException("More than one Delegation info in cert");
      	  else {
      	    delegateDefined = true;
      	    if (list.size() != 1)
      	      throw new SexpParseException("Delegation should be only a single word");
      	    else
      	      delegate = true;
      	  } 
      	}
      	else parseItem(list);
      }  
      else
      	throw new SexpParseException("Shouldn't have a non-list in a cert");
    }
    if (s == null || t == null) 
      throw new SexpParseException("Need subject and tag in cert");
  }
  public String getType() {
    return "Authorization Certificate";
  }
  public String toShortString() {
    return issuer.toShortString() + " -> " + s.toShortString();
  }
  public Tag getTag() {
    return t;
  }
}
