package sdsi;

import sdsi.sexp.*;

public class Def extends Cert {
  Name issuerName;
  
  
  /** MUST CHANGE BOTH CONSTRUCTORS SO WE CAN HAVE ONLY ONE ISSUER NAME*/
  /**Creates a new Def from an issuer name, a subject, and a Validity*/
  public Def(SDSIPrincipal issuer, String name, Subject s, Validity v) throws SDSIException{
    super(issuer,s,new Tag(),true,v);
    this.issuerName = new Name(issuer, new String[] {name});
    Sexp[] nameParts = new Sexp[2];
    nameParts[0] = new SexpString("issuer");
    nameParts[1] = issuerName.getSrep();
    Sexp[] subjParts = new Sexp[2];
    subjParts[0] = new SexpString("subject");
    subjParts[1] = s.getSrep();
    Sexp[] valids = v.getSreps();
    Sexp[] defParts = new Sexp[3 + valids.length];
    defParts[0] = new SexpString("cert");
    defParts[1] = new SexpList(nameParts);
    defParts[2] = new SexpList(subjParts);
    for (int i=0;i<valids.length;i++)
      defParts[3+i] = valids[i];
    srep = new SexpList(defParts);
  }
    
  /**Creates a new Def from a SexpList.  The Name which is the issuer 
    is also given, because in practice, this constructor is called from
    within Cert.create which has already parsed the issuer out of the 
    S-expression
    @exeption SexpParseException if the SexpList does not correctly represent
    a SDSI name defintion certificate*/
  public Def(SexpList l,Name issuerName) throws SexpParseException {
    super(l);
    delegate = true; //possibly need this here!!!!!!!!!!!
    if (issuerName.size() > 1)
      throw new SexpParseException("Only one name can be used as an issuer");
    this.issuerName = issuerName;
    this.issuer = issuerName.getPrincipal();
    this.t = new Tag();
    Sexp[] certParts = l.arrayOfElements(); 
    for (int j = 1; j < certParts.length; j++) {
      Sexp element = certParts[j];
      if (element instanceof SexpList) {
        SexpList list = (SexpList) element;
        String label = list.getType();
        parseItem(list);
      }
      else
	      throw new SexpParseException("Shouldn't have a non-list in a cert");
    }
    if (s == null) 
      throw new SexpParseException("Need subject in cert");
  }
  public String getType() {
    return "Name Certificate";
  }
  public String toShortString() {
    return issuerName.toShortString() + " => " + s.toShortString();
  }
  public Name getName() {
    return issuerName;
  }
}
