package sdsi;  //should extend SDSIObject????

import sdsi.sexp.*;
import java.security.*;

public class Hash extends SDSIPrincipal{
  /** Creates a new Hash from a S-expression list.
      @exception SexpParseException if the list is not a correctly
      formatted Hash.*/
  public Hash(SexpList l) throws SexpParseException {  //what do i want to do with s??????
    super(l);
    if (!l.getType().equals("hash"))
      throw new SexpParseException("Not a hash");
    if ((l.size() < 3) || !(l.elementAt(1) instanceof SexpString) || !(l.elementAt(2) instanceof SexpString))
      throw new SexpParseException("Incorrectly formatted Hash object");
    String alg = this.getHashID();
    byte[] value = this.getHashValue();
    if (alg.equals("md5")) {
      if (value.length != 16)
        throw new SexpParseException("Wrong hash length for MD5");
    }
    else if (alg.equals("sha1")) {
      if (value.length != 20)
        throw new SexpParseException("Wrong hash length for SHA-1");
    }
    else 
      throw new SexpParseException("Unrecognized hash algorithm");//should this be excep
  }
  
  /* do we want to create and store a canonical form for all new SDSIObjects we create
      if so then should hash be a SDSIObject*/
  //we should check for validity of algorithm and hash
  public Hash(String algorithm, byte[] value) { 
    Sexp[] list = new Sexp[3]; 
    list[0] = new SexpString("hash");
    list[1] = new SexpString(algorithm);
    list[2] = new SexpString(value);
    srep = new SexpList(list);
  }
  
  public Hash(String algorithm, SDSIObject s) throws NoSuchAlgorithmException{
    byte[] hash;
    MessageDigest md;
    Sexp[] list = new Sexp[3];
    list[0] = new SexpString("hash");
    list[1] = new SexpString(algorithm);
    if (algorithm.equals("md5")) {
      md = MessageDigest.getInstance("MD5");
      hash = s.getHash(md);
      list[2] = new SexpString(hash);
    }
    else if (algorithm.equals("sha1")) {
      md = MessageDigest.getInstance("SHA-1");
      hash = s.getHash(md);
      list[2] = new SexpString(hash);
    }
    else
      throw new NoSuchAlgorithmException();
    srep = new SexpList(list);
  }
   
  public String getHashID() { //do we want getAlgID and share Key interface? what else is in key interface
    return ((SexpString)srep.elementAt(1)).stringContent();
  }
  public byte[] getHashValue() { 
    return ((SexpString)srep.elementAt(2)).bytesContent(); 
  }
  public String getType() {
    return "Hash";
  }
  public String toShortString() {
    String val = srep.elementAt(2).toReadableString(0,10,0);
    return "Hash " + val.substring(0,(val.length() < 8) ? val.length() : 8);
  }
  public boolean equals(Object otherHash) {     //no maybe we want everything to be the same; change later
    //at least want algorithms to be the same!!
    if (otherHash instanceof Hash)
      return MessageDigest.isEqual(this.getHashValue(),((Hash)otherHash).getHashValue());
    else
      return false;
  }
  public int hashCode() {
    byte[] hash = this.getHashValue();
    int sum = 0;
    for (int i=0;i<hash.length;i++)
      sum += hash[i]*hash[i];
    return sum;
  }
} 

