package sdsi;

import java.security.*;
import sdsi.sexp.*;

public class Op extends SDSIObject {
  /** Creates a new Op from a S-expression list.
      @exception SexpParseException if the list is not a correctly
      formatted Op.*/
  public Op(SexpList l) throws SexpParseException{
    super(l);
    if (!l.getType().equals("do"))
      throw new SexpParseException("Not an Op");
    int numParts = l.size();
    if ((numParts < 2) || !(l.elementAt(1) instanceof SexpString))
      throw new SexpParseException("Incorrectly formatted Opcode");
    String opcode = this.getOpcode();
    if (opcode.equals("hash")) {
      if ((l.size() < 3) || !(l.elementAt(2) instanceof SexpString))
        throw new SexpParseException("Incorrectly formatted Hash Opcode");
    }
    else
      throw new SexpParseException("Not a recognized Op");
  }
  public String getOpcode() { 
    return ((SexpString)srep.elementAt(1)).stringContent();
  }
  public Hash execute(SDSIObject s) {
    byte[] hash;
    try {
      MessageDigest md;
      if (((SexpString)srep.elementAt(2)).stringContent().equals("md5"))
        md = MessageDigest.getInstance("MD5");
      else
        throw new NoSuchAlgorithmException();
        //md = null;
      hash = s.getHash(md);
    }
    catch (NoSuchAlgorithmException e) {
      System.out.println("That hash algorithm is not installed");
      hash = null;}
    return new Hash("md5",hash);
  } 
  public String getType() {
    return "Op";
  }
  public String toShortString() {//assumes its short enough
    return this.toReadableString(20);
  }
}
