package sdsi;

import sdsi.sexp.*;
import java.security.*;

public class SDSIPublicKey extends SDSIPrincipal implements PublicKey {
  
  SDSIPublicKey() {}
  
   /** Creates a new Public Key from a S-expression list.
      @exception SexpParseException if the list is not a correctly
      formatted Public Key.*/
  public SDSIPublicKey(SexpList l) throws SexpParseException{
    super(l);
     if (!l.getType().equals("public-key"))
      throw new SexpParseException("Not a public key");
  }
  
  public static SDSIPublicKey create(SexpList l) throws SexpParseException, SDSIException {
    if ((l.size() < 2) || !(l.elementAt(1) instanceof SexpList))
      throw new SexpParseException("Wrong format for SDSI Public Key");
    String alg = ((SexpList)l.elementAt(1)).getType(); 
    if ((alg.equals("rsa-pkcs1-md5")) || (alg.equals("rsa-pkcs1-sha1"))) 
      return new SDSIRSAPublicKey(l);
    else
      throw new SDSIException("Unknown Public-key algorithm");
  }
  
  //do we need getAlgID
  public String getAlgID() {  
    return "Unknown";
  }
  
  public String getAlgorithm() {
      return "Unknown";
  }
  
  public String getFormat() {  //we can implement these later if desired
    return null;
  }
  public byte[] getEncoded() {
    return null;
  }
  
  public String getURI() { //here we assume only one uri
    int numElements = srep.size();
    Sexp lastElement = srep.elementAt(numElements - 1);
    if (lastElement instanceof SexpString)
      return ((SexpString)lastElement).stringContent();
    else
      return null;
  }
  public String getType() {
    return "Public Key";
  }
  public boolean equals(Object other) {
    //later add a SexpList equals
    return false;
  }
  public boolean sameKeyAs(SDSIPrincipal p) {
    SDSIPublicKey other;
    if (p instanceof SDSIPublicKey) {
      return ((SDSIPublicKey)p).equals(this);
    }
    else { //p instanceof Hash
      try {
        Hash h = (Hash) p;
        Hash thisHash = new Hash(h.getHashID(),this);
        return h.equals(thisHash);
      }
      catch (NoSuchAlgorithmException e) {
        return false;
      }
    }
  }
}
