package sdsi;

import sdsi.sexp.*;
import java.security.interfaces.*;
import java.math.*;

public class SDSIRSAPrivateKey extends SDSIPrivateKey implements RSAPrivateKey {
  byte[] modulus;
  byte[] exponent; 
  
  public SDSIRSAPrivateKey(SexpList l) throws SexpParseException{
    super(l);
    String uri = this.getURI(); //allow multiple uri's
    boolean uriDefined = (uri == null) ? false : true;
    boolean eDefined = false, nDefined = false;
    Sexp[] keyParts = ((SexpList)l.elementAt(1)).arrayOfElements();
    for (int i = 1; i < keyParts.length; i++) { 
      //check for neg numbers?  //store numbers or store location of e and n??
      Sexp element = keyParts[i];
      if (element instanceof SexpList) {
        SexpList list = (SexpList) element;
        String label = list.getType();
        if (label.equals("e")) {
          if (eDefined)
            throw new SexpParseException("More than one exponent in RSA Public Key");
          else {
            eDefined = true;
            if ((list.size() != 2) || !(list.elementAt(1) instanceof SexpString))
              throw new SexpParseException("Exponent encoded incorrectly");
            exponent = ((SexpString)list.elementAt(1)).bytesContent();
          }
        }
        else if (label.equals("n")){
          if (nDefined)
            throw new SexpParseException("More than one modulus in RSA Public Key");
          else {
            nDefined = true;
            if ((list.size() != 2) || !(list.elementAt(1) instanceof SexpString))
              throw new SexpParseException("Modulus encoded incorrectly");
            modulus = ((SexpString)list.elementAt(1)).bytesContent();
          }
        }
        else
          //throw new SexpParseException("Unrecognized sub-type in RSA Public Key: " + label);
          ;//there maybe other subtypes in private keys
      }     
      else  
        throw new SexpParseException("Shouldn't have a non-list other than a URI in a Private Key");
    }
    if (!eDefined || !nDefined)
      throw new SexpParseException("Must define exponent and modulus in RSA Private Key");
  }
  public SDSIRSAPrivateKey(RSAPrivateKey k) {
    modulus = k.getModulus().toByteArray();
    exponent = k.getExponent().toByteArray();
    Sexp[] eParts = new Sexp[2];
    eParts[0] = new SexpString("e");
    eParts[1] = new SexpString(exponent);
    Sexp[] nParts = new Sexp[2];
    nParts[0] = new SexpString("n");
    nParts[1] = new SexpString(modulus);
    Sexp[] rsaParts = new Sexp[3];
    rsaParts[0] = new SexpString("rsa-pkcs1-md5");
    rsaParts[1] = new SexpList(eParts);
    rsaParts[2] = new SexpList(nParts);
    Sexp[] keyParts = new Sexp[2];
    keyParts[0] = new SexpString("private-key");
    keyParts[1] = new SexpList(rsaParts);
    
    srep = new SexpList(keyParts);
  }
  public String getAlgorithm() {
    String alg = ((SexpList)srep.elementAt(1)).getType();
    if (alg.equals("rsa-pkcs1-md5"))
      return "MD5/RSA";
    else if (alg.equals("rsa-pkcs1-sha1"))
      return "SHA-1/RSA";
    else 
      return "UNKNOWN";  
  }  
  public BigInteger getModulus() {
    return new BigInteger(modulus);
  }
  public BigInteger getExponent() {
    return new BigInteger(exponent);
  }
}  
