package sdsi;

import sdsi.sexp.*;
import java.security.interfaces.*;
import java.math.*;
import java.security.*;

public class SDSIRSAPublicKey extends SDSIPublicKey implements RSAPublicKey {
  byte[] modulus;
  byte[] exponent; 
  
  public SDSIRSAPublicKey(SexpList l) throws SexpParseException{
    super(l);
    String uri = this.getURI(); //change to allow multiple uri's//easier now though
    boolean uriDefined = (uri == null) ? false : true;
    boolean eDefined = false, nDefined = false;
    Sexp[] keyParts = ((SexpList)l.elementAt(1)).arrayOfElements();
    for (int i = 1; i < keyParts.length; i++) { 
      //check for neg numbers?  //store numbers or store location of e and n??
      Sexp element = keyParts[i];
      if (element instanceof SexpList) {
        SexpList list = (SexpList) element;
        String label = list.getType();
        if (label.equals("e")) {
          if (eDefined)
            throw new SexpParseException("More than one exponent in RSA Public Key");
          else {
            eDefined = true;
            if ((list.size() != 2) || !(list.elementAt(1) instanceof SexpString))
              throw new SexpParseException("Exponent encoded incorrectly");
            exponent = ((SexpString)list.elementAt(1)).bytesContent();
          }
        }
        else if (label.equals("n")){
          if (nDefined)
            throw new SexpParseException("More than one modulus in RSA Public Key");
          else {
            nDefined = true;
            if ((list.size() != 2) || !(list.elementAt(1) instanceof SexpString))
              throw new SexpParseException("Modulus encoded incorrectly");
            modulus = ((SexpString)list.elementAt(1)).bytesContent();
          }
        }
        else
          throw new SexpParseException("Unrecognized sub-type in RSA Public Key: " + label);
      }     
      else  
        throw new SexpParseException("Shouldn't have a non-list other than a URI in a Public Key");
    }
    if (!eDefined || !nDefined)
      throw new SexpParseException("Must define exponent and modulus in RSA Public Key");
  }
  public SDSIRSAPublicKey(RSAPublicKey k) {
    modulus = k.getModulus().toByteArray();
    exponent = k.getExponent().toByteArray();
    Sexp[] eParts = new Sexp[2];
    eParts[0] = new SexpString("e");
    eParts[1] = new SexpString(exponent);
    Sexp[] nParts = new Sexp[2];
    nParts[0] = new SexpString("n");
    nParts[1] = new SexpString(modulus);
    Sexp[] rsaParts = new Sexp[3];
    rsaParts[0] = new SexpString("rsa-pkcs1-md5");
    rsaParts[1] = new SexpList(eParts);
    rsaParts[2] = new SexpList(nParts);
    Sexp[] keyParts = new Sexp[2];
    keyParts[0] = new SexpString("public-key");
    keyParts[1] = new SexpList(rsaParts);
    
    srep = new SexpList(keyParts);
  }
  public String getAlgorithm() {
    String alg = ((SexpList)srep.elementAt(1)).getType();
    if (alg.equals("rsa-pkcs1-md5"))
      return "MD5/RSA";
    else if (alg.equals("rsa-pkcs1-sha1"))
      return "SHA-1/RSA";
    else 
      return "UNKNOWN";
  }  
  public BigInteger getModulus() {
    return new BigInteger(modulus);
  }
  public BigInteger getExponent() {
    return new BigInteger(exponent);
  }
  public String toShortString() {
    String num = new BigInteger(modulus).toString(16);
    return "K-" + num.substring(0,(num.length() < 6) ? num.length() : 6);
  }
  public boolean equals(Object other) {
    if (other instanceof SDSIRSAPublicKey) {
      return (MessageDigest.isEqual(this.modulus,((SDSIRSAPublicKey)other).modulus) &&
             MessageDigest.isEqual(this.exponent,((SDSIRSAPublicKey)other).exponent));
    }
    else
      return false;
  }
  public int hashCode() {
    int sum = 0;
    for (int i = 0;i<modulus.length;i++)
      sum += modulus[i]*modulus[i];
    for (int i = 0;i<exponent.length;i++)
      sum += exponent[i]*exponent[i];
    return sum;
  }
}  
