package sdsi;  //should extend SDSIObject????

import sdsi.sexp.*;
import java.security.*;
import java.math.*;
import java.util.*;

public class SDSISignature extends SDSIObject {
  SDSIPrincipal principal;
  Hash hashOfObject;
  
  public SDSISignature(SDSIObject signedObject, 
                       SDSIPrivateKey privKey, 
                       SDSIPublicKey pubKey) throws SignatureException{
    try {
      this.principal = pubKey; 
      this.hashOfObject = new Hash("md5",signedObject);
      Signature privSign = Signature.getInstance(privKey.getAlgorithm());
      privSign.initSign(privKey);
      privSign.update(signedObject.getSrep().getCanonRep());
      byte[] signature = privSign.sign();
      Sexp[] sigParts = new Sexp[4];
      sigParts[0] = new SexpString("signature");
      sigParts[1] = hashOfObject.getSrep();
      sigParts[2] = pubKey.getSrep();
      sigParts[3] = new SexpString(signature);
      srep = new SexpList(sigParts);
    }
    catch (NoSuchAlgorithmException e) {//won't happen
      throw new SignatureException("No MD5 hash algorithm");
    }
    catch (InvalidKeyException e) {
      throw new SignatureException(e.getMessage());
    }
  }
  public SDSISignature(SexpList l) throws SexpParseException{
    super(l);
    if (!l.getType().equals("signature"))
      throw new SexpParseException("Not a signature");
    if ((l.size() < 4))
      throw new SexpParseException("Incorrectly formatted Signature object: too few parts");
    if (!(srep.elementAt(1) instanceof SexpList) || !(srep.elementAt(2) instanceof SexpList) 
                                                 || !(srep.elementAt(3) instanceof SexpString)) //can sigval be list?
      throw new SexpParseException("Incorrectly formatted Signature: parts of wrong type");
    SDSIObject supposedHash = SDSIObject.principalParse((SexpList)srep.elementAt(1),null);
    if (!(supposedHash instanceof Hash))
      throw new SexpParseException("Incorrectly formatted Signature: first part wasn't a hash");
    else
      hashOfObject = (Hash)supposedHash;
    SDSIObject supposedPrincipal = SDSIObject.principalParse((SexpList)srep.elementAt(2),null);
    if (!(supposedPrincipal instanceof SDSIPrincipal))
      throw new SexpParseException("Incorrectly formatted Signature: principal must be Public Key or hash of");
    else
      principal = (SDSIPrincipal)supposedPrincipal;
  }
  /*
  public void setSignedObject(SDSIObject o) {
    signedObject = o;
  }*/
  public String toShortString() {
    return "Signature";
  }
  public String getType() {
    return "Signature";
  }
  public SDSIPrincipal getPrincipal() {
    return principal;
  }
 /* 
  public SDSIObject getSignedObject() {
    return signedObject;
  }*/
  public Hash getHashOfObject() {
    return hashOfObject;
  }
  public byte[] getSignatureValue() { //need to worry about display type???
    return ((SexpString)srep.elementAt(3)).bytesContent(); 
  }
  public boolean verify(SDSIPublicKey verifyKey, SDSIObject objToVerify) throws SexpParseException, NoSuchAlgorithmException,
                                            SignatureException, InvalidKeyException {
    //watch for cast class exceptions when using SexpString/SexpList
    byte[] sig = this.getSignatureValue();
    Signature verifySign = Signature.getInstance(verifyKey.getAlgorithm());
    verifySign.initVerify(verifyKey);
    objToVerify.sign(verifySign);
    //Check keys are the same
    if (principal instanceof SDSIPublicKey) {
      if (!(verifyKey.equals((SDSIPublicKey)principal)))
        throw new InvalidKeyException("Key given doesn't match key in signature");
    }
    else { //principal instanceof Hash
      Hash hashPubKey = new Hash(((Hash)principal).getHashID(),verifyKey);
      if (!((Hash)principal).equals(hashPubKey))
        throw new InvalidKeyException("Key given doesn't match key in signature");
    }
    //Checking Hash
    Hash HashObjToVerify = new Hash(hashOfObject.getHashID(),objToVerify);
    if (hashOfObject.equals(HashObjToVerify))
      return verifySign.verify(sig);
    else
      throw new SignatureException("Hash in signature doesn't match hash of object"); //might want to treat this differently
  }
}
