package sdsi;  //should extend SDSIObject????

import sdsi.sexp.*;
import java.io.*;

public class Tag extends SDSIObject {
  TagExpression t;  //may not want to carry this around 
  //may just want to create it when needed and once at the beginning to check formatting
  public Tag() {//the (*) tag with no S-exp representation
    t = new TagStar();
    srep = new SexpList(new Sexp[] {new SexpString("tag"),
                                    new SexpList(new Sexp[] {new SexpString("*")})}); 
  }
  public Tag(SexpList l) throws SexpParseException{
    super(l);
    if (!(l.size() == 2)) 
      throw new SexpParseException("Incorrectly formatted tag");
    SexpList pTagStar;
    if (l.elementAt(1) instanceof SexpList 
       && (pTagStar = ((SexpList)l.elementAt(1))).size() == 1 
       && pTagStar.elementAt(0) instanceof SexpString 
       && ((SexpString)pTagStar.elementAt(0)).stringContent().equals("*"))
      t = new TagStar();
    else
      t = TagExpression.parse(l.elementAt(1));
  }
  public String toShortString() {
    String s = srep.toReadableString(0,200,0);
    return s.substring(5,(s.length() < 21) ? s.length()-1 : 20);
  }
  public boolean intersect(Tag starFree) { //don't really like this
    TagExpression sf = starFree.t;
    if (t.intersect(sf) != null)
      return true;
    else
      return false;
  }
  public abstract static class TagExpression implements Serializable{
    public static TagExpression parse(Sexp s) throws SexpParseException{
      if (s instanceof SexpString)
        return new TagByteString((SexpString)s);
      else { //s is a SexpList
        SexpList l = (SexpList) s;
        if (l.elementAt(0) instanceof SexpList)
          throw new SexpParseException("First element in tag can't be list");
        if (((SexpString)l.elementAt(0)).stringContent().equals("*")) {
          return StarTag.parse(l);
        }
        else
          return new SimpleTag(l);
      }
    }
    public abstract TagExpression intersect(TagExpression starFree);
  }
  
  public static class SimpleTag extends TagExpression{
    /*  so we don't allow tags that look like
    ((* set read write) file foo) ?
    */
    SexpString first;
    TagExpression[] rest;
    public SimpleTag(SexpList l) throws SexpParseException {
      int size = l.size();
      first = (SexpString)l.elementAt(0);
      rest = new TagExpression[size - 1];
      for (int i =0;i<size-1;i++)
        rest[i] = TagExpression.parse(l.elementAt(i+1));
    }
    public TagExpression intersect(TagExpression starFree) {
      if (starFree instanceof StarTag)
        return null;
      else {
        if (starFree instanceof SimpleTag) {
          SimpleTag simpTag = (SimpleTag)starFree;
          if (simpTag.rest.length != this.rest.length) //require lengths equal
            return null;
          if (simpTag.first.equals(this.first)) {
            for (int i = 0;i<this.rest.length;i++) {
              if ((simpTag.rest[i] = this.rest[i].intersect(simpTag.rest[i])) == null)
                return null;
            }
            return simpTag;
          }
          else
            return null;
        }
        else
          return null;
      }
    }
  }
  
  public abstract static class StarTag extends TagExpression{
    public static StarTag parse(SexpList l) throws SexpParseException{
      if (l.size() < 2 || !(l.elementAt(1) instanceof SexpString))
        throw new SexpParseException("Tag with star not encoded properly");
      String starType = ((SexpString)l.elementAt(1)).stringContent();
      if (starType.equals("set"))
        return new TagSet(l);
      else if (starType.equals("range"))
        return new TagRange(l);
      else if (starType.equals("prefix"))
        return new TagPrefix(l);
      else
        throw new SexpParseException("Unknown star type in tag");
    }
  }
  
  public static class TagStar extends StarTag{
    public TagStar() {;}
    public TagExpression intersect(TagExpression starFree) {
      if (starFree instanceof StarTag)
        return null;
      else 
        return starFree;
    }
  }
  
  public static class TagRange extends StarTag{
    boolean stringCompare = false;
    boolean strictlyGreater = false;
    boolean strictlyLess = false;
    SexpString lowerLimit = null;  //can store these as strings if we're not carrying them everywhere
    SexpString upperLimit = null;  //same in prefix
    public TagRange(SexpList l) throws SexpParseException { 
      int size = l.size();
      if (!(size == 3 || size ==5 || size == 7)) 
        throw new SexpParseException("*-range type tag not encoded properly");
      for (int i=2;i<size;i++) {
        if (!(l.elementAt(i) instanceof SexpString))
          throw new SexpParseException("*-range type tag not encoded properly");
      }  
      String rangeType = ((SexpString)l.elementAt(2)).stringContent();
      if (rangeType.equals("alpha") || rangeType.equals("numeric") || rangeType.equals("time")
                                    || rangeType.equals("binary") || rangeType.equals("date")) {
        stringCompare = true;
      }    
      else
        throw new SexpParseException("Unknown range type in tag");
      if (size == 5) {
        String strictness = ((SexpString)l.elementAt(3)).stringContent();
        if (strictness.equals("l")) {
          strictlyLess = true;
          upperLimit = (SexpString)l.elementAt(4);
        }
        else if (strictness.equals("le")) {
          strictlyLess = false;
          upperLimit = (SexpString)l.elementAt(4);
        }
        else if (strictness.equals("g")) {
          strictlyGreater = true;
          lowerLimit = (SexpString)l.elementAt(4);
        }
        else if (strictness.equals("ge")) {
          strictlyGreater = false;
          lowerLimit = (SexpString)l.elementAt(4);
        }
        else 
          throw new SexpParseException("*-range type tag not encoded properly");
      }
      if (size == 7) {
        String strictness = ((SexpString)l.elementAt(3)).stringContent();
        if (strictness.equals("g")) 
          strictlyGreater = true;
        else if (strictness.equals("ge")) 
          strictlyGreater = false;
        else 
          throw new SexpParseException("*-range type tag not encoded properly");
        strictness = ((SexpString)l.elementAt(5)).stringContent();
        if (strictness.equals("l")) 
          strictlyLess = true;
        else if (strictness.equals("le")) 
          strictlyLess = false;
        else 
          throw new SexpParseException("*-range type tag not encoded properly");
        lowerLimit = (SexpString)l.elementAt(4);
        upperLimit = (SexpString)l.elementAt(6);
      }
    }
    public TagExpression intersect(TagExpression starFree) {
      if (starFree instanceof StarTag)
        return null;
      else {
        if (!(starFree instanceof TagByteString))
          return null;
        SexpString starFreeString = ((TagByteString)starFree).value;
        try {
          if (upperLimit != null) { //no upper and lower limit means all range
            if (strictlyLess) {
              if (stringCompare)  //may want to move these??
                if (starFreeString.compareTo(upperLimit) >= 0)
                  return null;
              else
                return null;
            }
            else {
              if (stringCompare)
                if (starFreeString.compareTo(upperLimit) > 0)
                  return null;
              else
                return null;
            }
          }
          if (lowerLimit != null) {
            if (strictlyGreater) {
              if (stringCompare)
                if (starFreeString.compareTo(lowerLimit) <= 0)
                  return null;
              else
                return null;
            }
            else {
              if (stringCompare)
                if (starFreeString.compareTo(upperLimit) < 0)
                  return null;
              else 
                return null;
            }
          }
         return starFree;
        }
        catch (SexpException e) {
          return null;
        }
      }
    }
  }
  
  public static class TagSet extends StarTag{
    TagExpression[] components;
    public TagSet(SexpList l) throws SexpParseException{
      int size = l.size();
      components = new TagExpression[size - 2];
      for (int i =0;i<size-2;i++)
        components[i] = TagExpression.parse(l.elementAt(i+2));
    }
    public TagExpression intersect(TagExpression starFree) {
      if (starFree instanceof StarTag)
        return null;
      else {
        for (int i = 0;i<components.length;i++) {
          if (components[i].intersect(starFree) != null)  //this won't work unless intersection returns whole starfree
            return starFree;
        }
        return null;
      }
    }
  }
  
  public static class TagPrefix extends StarTag {
    SexpString prefix;
    public TagPrefix(SexpList l) throws SexpParseException{
      if (l.size() != 3 || !(l.elementAt(2) instanceof SexpString))
        throw new SexpParseException("Tag with *-prefix not encoded properly");
      prefix = (SexpString)l.elementAt(2);
    }  
    public TagExpression intersect(TagExpression starFree) {
      if (starFree instanceof StarTag)
        return null;
      else {
        if (!(starFree instanceof TagByteString))
          return null;
        SexpString starFreeString = ((TagByteString)starFree).value;
        if (starFreeString.startsWith(prefix))
          return starFree;
        else
          return null;
      }
    }   
  }
  
  public static class TagByteString extends TagExpression {
    SexpString value;
    public TagByteString(SexpString s) {
      value = s;
    }
    public TagExpression intersect(TagExpression starFree) {
      if (starFree instanceof StarTag)
        return null;
      else {
        if (!(starFree instanceof TagByteString))
          return null;
        if (this.value.equals(((TagByteString)starFree).value))
          return starFree;
        else 
          return null;
      }
    }
  }
}
