package sdsi;

public class Tuple {
  
  boolean namingPermission;
  
  Validity v;
  boolean propagate;
  boolean toplevel;
  Tag t;
  SDSIPublicKey subject;
  Name subjectName;
  SDSIPublicKey issuer;
  Name issuerName;
  
  public Tuple(Cert c, Tag request, SDSIPublicKey subject, SDSIPublicKey issuer) throws VerificationException {
    this.v = c.valid;
    this.propagate = c.delegate;
    this.subject = subject;
    this.issuer = issuer;
    if (c.t.intersect(request))
      this.t = request;
    else
      this.t = null;
    if (c.s instanceof Name)
      subjectName = (Name)c.s;
    else
      subjectName = null;
    if (this.t == null) 
      throw new VerificationException("Certificate tag does not match request");
    if (c instanceof Def) {
      namingPermission = true;
      toplevel = false;
      this.issuerName = ((Def)c).issuerName;
    }
    else {
      namingPermission = false;
      toplevel = true;
      this.issuerName = null;
    } 
  }
  
  public void add(Tuple p) throws VerificationException {
    v.update(p.v);
    if (!p.t.intersect(this.t)) 
      throw new VerificationException("Certificate tag does not match request");
    /*Note at this point p will only have one name where as this may have more than one*/
    /*Named subsequences may change this*/
    if (!p.issuer.equals(subject))
      throw new VerificationException("Cert doesn't follow: " + p.issuer.toShortString());
    if (p.namingPermission) {
      if (subjectName == null /*|| this.subjectName.prefixNamesOf(p.issuerName)*/) {
        if (toplevel)
          throw new VerificationException("Already at the top level, can't expand subject");
        else {
          issuerName.expandNames(subjectName,p.issuerName);
          subjectName = p.subjectName;
        }
      }
      else if (subjectName.sameNamesAs(p.issuerName)) {
        subjectName = p.subjectName;
      }
      else if (p.issuerName.prefixNamesOf(subjectName)) {
        subjectName.replaceNames(p.issuerName,p.subjectName,p.subject);
      }
      else 
        throw new VerificationException("Names don't match up");
    }
    else {
      toplevel = true;
      if (this.subjectName != null)
        throw new VerificationException("Can't have an auth, except at top level");
      else
        this.subjectName = p.subjectName;
    }
    this.subject = p.subject;
  }
  public boolean valid() {
    return v.valid();
  }
  public String toReadableString(int width) {
    StringBuffer answer = new StringBuffer("");
    answer.append("ISSUER\n"); 
    if (issuerName != null)
      answer.append(issuerName.toReadableString(width));
    else
      answer.append(issuer.toReadableString(width));
    answer.append("\n\nSUBJECT\n");
    if (subjectName != null)
      answer.append(subjectName.toReadableString(width));
    else
      answer.append(subject.toReadableString(width));
    answer.append("\n\nTAG\n" + t.toReadableString(width));
    return answer.toString();
  }
    
}
    
    