package sdsi;

import sdsi.sexp.*;
import java.util.*;
import sdsi.control.*;  //maybe move certcache to the sdsi package


public class Verification {
  
  CertCache cache; //just key hashes for now
  Acl acl;
  Sequence sequence;
  Tag request;
  SDSISignature signatureOnRequest;
  Tuple result;
 
  public Verification(CertCache cache) {
    this.cache = cache;
  }
  public void initializeAcl(Acl acl) {
    this.acl = acl;
  }
  public void setRequest(Tag request) {
    this.request = request;
  }
  public void setSequence(Sequence sequence) {
    this.sequence = sequence;
  }
  public void setSignature(SDSISignature signatureOnRequest) {
    this.signatureOnRequest = signatureOnRequest; //hmm
  }

  public String verifyAuthority() throws VerificationException{
    try {
      Tuple granted = processSequence();
      result = granted;
      SDSIPublicKey grantor = granted.issuer;
      if (signatureOnRequest == null) 
        return "No signature to check against";
      if (acl == null)
        return "No acl to check against";
      SDSIPrincipal requestIssuer = signatureOnRequest.getPrincipal();
      SDSIPublicKey requestor;
      if (requestIssuer instanceof SDSIPublicKey)
        requestor = (SDSIPublicKey) requestIssuer;
      else { //requestIssuer instanceof Hash
         SDSIObject o = (SDSIObject)cache.get((Hash)requestIssuer,CertCache.KEY);
        if (o == null || !(o instanceof SDSIPublicKey))
          return "Error with request signature";
        requestor = (SDSIPublicKey) o;
      }
      if (signatureOnRequest.verify(requestor,request)) {
        if (granted.valid()) {
          if (granted.subjectName == null && granted.subject.equals(requestor)) {
            if (acl.contains(granted.issuer, granted.issuerName, request))
              return "Verification Authorized";
            else 
              return "ACL does not contain a matching entry";
          }
          else
            return "Sequence does not reduce to issuer of request";
        }
        else 
          return "Sequence is no longer valid";
      }
      else
        return "Signature does not match request";
    }
    catch (Exception e) {
      throw new VerificationException("Exception during verification: " + e.getMessage());
    }
  }
  public Tuple processSequence() throws SDSIException, VerificationException {
    if (request == null || sequence == null)
      throw new VerificationException("Must submit request and sequence first");
    SDSIObject[] sequencePart = sequence.members();
    Tuple current = null;
    Tuple next = null;
    SDSIPublicKey subject;
    SDSIPublicKey issuer = null;
    Cert currentCert = null;
    boolean lastCertNeedsSig = false;
    try {
      for (int i=0;i<sequencePart.length; i++) {
        if (sequencePart[i] instanceof Cert) {
          if (lastCertNeedsSig)
            throw new SDSIException("Each Cert Must be Signed");
          else { //right now only accept certs in order and only accept one sig per cert
            currentCert = (Cert)sequencePart[i];
            Subject certSubject = currentCert.s;
            SDSIPrincipal certIssuer = currentCert.issuer;
            if (certSubject instanceof Name) 
              certSubject = ((Name)certSubject).getPrincipal();
            if (certSubject instanceof SDSIPublicKey)
              subject = (SDSIPublicKey) certSubject;
            else if (certSubject instanceof Hash) {
              SDSIObject o = (SDSIObject)cache.get((Hash)certSubject,CertCache.KEY);
              if (o == null || !(o instanceof SDSIPublicKey))
                throw new VerificationException("Can't find subject Hash, or hash isn't of Public Key");
              subject = (SDSIPublicKey) o;
            }
            else
              throw new VerificationException("Subject of Cert should be PublicKey or hash");
            if (certIssuer instanceof SDSIPublicKey)
              issuer = (SDSIPublicKey) certIssuer;
            else {// (certIssuer instanceof Hash)
              SDSIObject o = (SDSIObject)cache.get((Hash)certIssuer,CertCache.KEY);
              if (o == null || !(o instanceof SDSIPublicKey))
                throw new VerificationException("Can't find issuer Hash, or hash isn't of Public Key");
              issuer = (SDSIPublicKey) o;
            } 
            next = new Tuple(currentCert,request,subject,issuer);
            lastCertNeedsSig = true;
          }
        }
        else if (sequencePart[i] instanceof SDSISignature) {
          if (lastCertNeedsSig) {
            SDSISignature signatureToVerify = (SDSISignature) sequencePart[i];
            if (signatureToVerify.verify(issuer,currentCert)) {
              lastCertNeedsSig = false;
              cache.add(currentCert,signatureToVerify);
              if (current == null)
               current = next;
              else
               current.add(next);
            }
            else
             throw new VerificationException("Signature does not verify on Cert");
          }
        }
        else if (sequencePart[i] instanceof SDSIPublicKey) {
          cache.add(sequencePart[i]);
        }
        else if (sequencePart[i] instanceof Op) {
          if (((Op)sequencePart[i]).getOpcode().equals("hash")) {
            if (i==0)
              throw new VerificationException("Sequence can't start with hash op");
            Hash h = ((Op)sequencePart[i]).execute(sequencePart[i-1]);
            //check to see if given hash is accurate
            cache.add(sequencePart[i-1]); //add as all hashes
          }
        }
      }
      return current;
    }
    catch (Exception e) {
      throw new VerificationException("Error during sequence processing: " + e.getMessage());
    } 
  }
  public String getReduction(int width) {
    if (result == null) 
      return "Not processed";
    else
      return result.toReadableString(width);
  }
}


