package sdsi.control;

import java.util.*;
import java.io.*;
import java.security.*;
import sdsi.*;
import sdsi.sexp.*;

public class CertCache implements Serializable{
  
  SDSIPublicKey owner;
  SDSIPrivateKey ownerPrivate;
  
  private static final int definedHashTypes = 2;
  
  transient Hashtable displayedObjects;
  
  int currentStored;
  
  Hashtable myAuth;
  Hashtable myDef;
  Hashtable otherAuth;
  Hashtable otherDef;
  Hashtable allKey;
  Hashtable miscObj;
  
  Hashtable hashMyAuth;
  Hashtable hashMyDef;
  Hashtable hashOtherAuth;
  Hashtable hashOtherDef;
  Hashtable hashAllKey;
  Hashtable hashMiscObj;
  
  Hashtable keyNames;
  
  /*later add functionality to load a previously saved certcache
  eventually add functionality to use the cache as a file and only pull
  into memory as neede (possibly one hashtable at a time)*/
  public CertCache(SDSIPublicKey owner, SDSIPrivateKey ownerPrivate) {
    this.owner = owner;
    this.ownerPrivate = ownerPrivate;
    myAuth = new Hashtable();
    myDef = new Hashtable();
    otherAuth = new Hashtable();
    otherDef = new Hashtable();
    allKey = new Hashtable();
    miscObj = new Hashtable();
    hashMyAuth = new Hashtable();
    hashMyDef = new Hashtable();
    hashOtherAuth = new Hashtable();
    hashOtherDef = new Hashtable();
    hashAllKey = new Hashtable();
    hashMiscObj = new Hashtable();
    displayedObjects = new Hashtable();
    keyNames = new Hashtable();
    currentStored = 0;
  }
  private void readObject(ObjectInputStream in) throws IOException, ClassNotFoundException{
    in.defaultReadObject();
    this.displayedObjects = new Hashtable();
  }
  public void add(Cert cert, SDSISignature sig) {
    int i;
    Hash[] hashTypes = getAllHashes(cert);
    CertSig cs = new CertSig(cert,sig);
    Integer index = new Integer(currentStored++);
    Integer oldIndex = null;
    
    if (sameKeys(owner, cert.getIssuer())) {
      if (cert instanceof Def) {
        for (i = 0;i<hashTypes.length;i++) 
          oldIndex = (Integer)hashMyDef.put(hashTypes[i],index);
        myDef.put(index,cs);
        if (oldIndex != null)
          myDef.remove(oldIndex);
        updateKeyNames((Def)cert);
      }
      else {
        for (i = 0;i<hashTypes.length;i++) 
          oldIndex = (Integer)hashMyAuth.put(hashTypes[i],index);
        myAuth.put(index,cs);
        if (oldIndex != null)
          myAuth.remove(oldIndex);
      }
    }
    else {
      if (cert instanceof Def) {
        for (i = 0;i<hashTypes.length;i++) 
          oldIndex = (Integer)hashOtherDef.put(hashTypes[i],index);
        otherDef.put(index,cs);
        if (oldIndex != null)
          otherDef.remove(oldIndex);
        updateKeyNames((Def)cert);
      }
      else {
        for (i = 0;i<hashTypes.length;i++) 
          oldIndex = (Integer)hashOtherAuth.put(hashTypes[i],index);
        otherAuth.put(index,cs);
        if (oldIndex != null)
          otherAuth.remove(oldIndex);
      }
    }
  }
  
  public void add(SDSIObject obj) {
    Hash[] hashTypes = getAllHashes(obj);
    int i;
    Integer index = new Integer(currentStored++);
    Integer oldIndex = null;
    
    if (obj instanceof SDSIPublicKey) {
      for (i = 0;i<hashTypes.length;i++) 
          oldIndex = (Integer)hashAllKey.put(hashTypes[i],index);
      allKey.put(index,obj);
      if (oldIndex != null)
          allKey.remove(oldIndex);
    }
    else {
      for (i = 0;i<hashTypes.length;i++) 
          oldIndex = (Integer)hashMiscObj.put(hashTypes[i],index);
      miscObj.put(index,obj);
      if (oldIndex != null)
          miscObj.remove(oldIndex);
    }
  }
  
  public void removeMisc(SDSIObject obj) {   
    int i;
    Integer index = null;
    Hash[] hashTypes = getAllHashes(obj);
    for (i = 0;i<hashTypes.length;i++) 
      index = (Integer)hashMiscObj.remove(hashTypes[i]);
    if (index != null)
      miscObj.remove(index);
  }
  
  public void removeAll(SDSIObject obj) {
    Hash[] hashTypes = getAllHashes(obj);
    Integer index = null;
    int i;
    if (obj instanceof Cert) {
      Cert cert = (Cert) obj;
      if (sameKeys(owner, cert.getIssuer())) {
        if (cert instanceof Def) {
          for (i = 0;i<hashTypes.length;i++) 
            index = (Integer)hashMyDef.remove(hashTypes[i]);
          if (index != null)
            myDef.remove(index);
          removeFromKeyNames((Def)cert);
        }
        else {
          for (i = 0;i<hashTypes.length;i++) 
            index = (Integer)hashMyAuth.remove(hashTypes[i]);
          if (index != null)
            myAuth.remove(index);
        }
      }
      else {
        if (cert instanceof Def) {
          for (i = 0;i<hashTypes.length;i++) 
            index = (Integer)hashOtherDef.remove(hashTypes[i]);
          if (index != null)
            otherDef.remove(index);
          removeFromKeyNames((Def)cert);
        }
        else {
          for (i = 0;i<hashTypes.length;i++) 
            index = (Integer)hashOtherAuth.remove(hashTypes[i]);
          if (index != null)
            otherAuth.remove(index);
        }
      }
    }
    else if (obj instanceof SDSIPublicKey) {
      for (i = 0;i<hashTypes.length;i++) 
        index = (Integer)hashAllKey.remove(hashTypes[i]);
      if (index != null)
        allKey.remove(index);
    }
    for (i = 0;i<hashTypes.length;i++) 
      index = (Integer)hashMiscObj.remove(hashTypes[i]);
    if (index != null)
      miscObj.remove(index);
  }
  
  public CertSig getBoth(Hash h, int type) {
    Integer index = null;
    switch (type) {
    case AUTH:
      index = (Integer)hashMyAuth.get(h);
      if (index != null)
        return (CertSig)myAuth.get(index);
      else if ((index = (Integer)hashOtherAuth.get(h)) != null)
        return (CertSig)otherAuth.get(index);
      break;
    case DEF:
      index = (Integer)hashMyDef.get(h);
      if (index != null)
        return (CertSig)myDef.get(index);
      else if ((index = (Integer)hashOtherDef.get(h)) != null)
        return (CertSig)otherDef.get(index);
      break;
    case UNKNOWN:
      CertSig cs;
      if ((cs = getBoth(h,AUTH)) != null)
        return cs;
      else
        return getBoth(h,DEF);
    }
    return null;
  }
  
  public SDSIObject get(Hash h, int type) {
    CertSig cs;
    Integer index = null;
    switch (type) {
    case KEY:
      if ((index = (Integer)hashAllKey.get(h)) != null)
        return (SDSIObject)allKey.get(index);
      else
        break;
    case AUTH:
      cs = getBoth(h,AUTH);
      if (cs != null)
        return cs.cert;
      else
        break;
    case DEF:
      cs = getBoth(h,DEF);
      if (cs != null)
        return cs.cert;
      else
        break;
    case OTHER:
      if ((index = (Integer)hashMiscObj.get(h)) != null)
        return (SDSIObject)miscObj.get(index);
      else
        break;
    case UNKNOWN:
      if ((cs = getBoth(h,UNKNOWN)) != null)
        return cs.cert;
      else if ((index = (Integer)hashAllKey.get(h)) != null)
        return (SDSIObject)allKey.get(index);
      else if ((index = (Integer)hashMiscObj.get(h)) != null)
        return (SDSIObject)miscObj.get(index);
      else
        break;
    }
    return null;
  }
  
  public SDSIPublicKey getOwner() {
    return owner;
  }
  public SDSIPrivateKey getOwnerPrivate() {
    return ownerPrivate;
  }
  
  public void setDisplay(Class c) {
    displayedObjects.clear();
    String s = c.getName();
    if (s.equals("sdsi.Auth")){
      addAuthsToDisplay(true);
      addAuthsToDisplay(false);
    }
    else if (s.equals("sdsi.Def")) {
      addDefsToDisplay(true);
      addDefsToDisplay(false);
    }
    else if (s.equals("sdsi.SDSIPublicKey")) {
      addKeysToDisplay();
    }
    else {
      Integer index;
      SDSIObject o;
      for (Enumeration objs = miscObj.keys();objs.hasMoreElements();) {
        index = (Integer)objs.nextElement();
        o = (SDSIObject)miscObj.get(index);
        if (c.isAssignableFrom(o.getClass())) {
          displayedObjects.put(getDisplayName(o, index), o);
        }
      }
      if (s.equals("sdsi.SDSIObject")) {
        addAuthsToDisplay(true);
        addAuthsToDisplay(false);
        addDefsToDisplay(true);
        addDefsToDisplay(false);
        addKeysToDisplay();
      }
    }
  }
  
  public void setDisplayLimited(Class c, boolean mine) {
    displayedObjects.clear();
    String s = c.getName();
    if (s.equals("sdsi.Auth")) {
      addAuthsToDisplay(mine);
    }
    else if (s.equals("sdsi.Def")) {
      addDefsToDisplay(mine);
    }
    else
      setDisplay(c);
  }
  private void addAuthsToDisplay(boolean mine) {
    if (mine) {
      for (Enumeration objs = myAuth.keys();objs.hasMoreElements();) {
        Integer i = (Integer)objs.nextElement();
        CertSig c = (CertSig)myAuth.get(i);
        displayedObjects.put(getDisplayName(c.cert,i), c);
      }
    }
    else {
      for (Enumeration objs = otherAuth.keys();objs.hasMoreElements();) {
        Integer i = (Integer)objs.nextElement();
        CertSig c = (CertSig)otherAuth.get(i);
        displayedObjects.put(getDisplayName(c.cert,i), c);
      }
    }
  }
  private void addDefsToDisplay(boolean mine) {
    if (mine) {
      for (Enumeration objs = myDef.keys();objs.hasMoreElements();) {
        Integer i = (Integer)objs.nextElement();
        CertSig c = (CertSig)myDef.get(i);
        displayedObjects.put(getDisplayName(c.cert,i), c);
      }
    }
    else {
      for (Enumeration objs = otherDef.keys();objs.hasMoreElements();) {
        Integer i = (Integer)objs.nextElement();
        CertSig c = (CertSig)otherDef.get(i);
        displayedObjects.put(getDisplayName(c.cert,i), c);
      }
    }
  }
  private void addKeysToDisplay() {
    for (Enumeration objs = allKey.keys();objs.hasMoreElements();) {
      Integer i = (Integer)objs.nextElement();
      SDSIPublicKey k = (SDSIPublicKey)allKey.get(i);
      displayedObjects.put(getDisplayName(k,i), k);
    }
  }
  private String getDisplayName(SDSIObject o,Integer index) {
    StringBuffer name = new StringBuffer("<" + index.toString() + ">  ");
    if (o instanceof Hash) {
      name.append("[Hash] ");
      o = get((Hash)o,UNKNOWN);
      if (o == null) {
        name.append("Unknown");
        return name.toString();
      }
    }
    if (o instanceof Def) {
      Name n = ((Def)o).getName();
      if (!sameKeys(owner, n.getPrincipal())) { 
        name.append(getNameFor(n.getPrincipal()) + "'s ");
      }
      name.append(n.getNameIndex(0));
      for (int i=1; i<n.size();i++) {
        name.append("'s " + n.getNameIndex(i));
      }
    }
    else if (o instanceof Auth) {
      SDSIPrincipal issuer = ((Auth)o).getIssuer();
      if (!sameKeys(owner,issuer))
        name.append(getNameFor(issuer));
      name.append(" #" + ((Auth)o).getTag().toShortString());
      Subject s = ((Auth)o).getSubject();
      name.append("# ");
      if (s instanceof SDSIPrincipal)
        name.append(getNameFor((SDSIPrincipal)s));
      else if (s instanceof Name) {
        if (!sameKeys(owner, ((Name)s).getPrincipal())) { 
          name.append(getNameFor(((Name)s).getPrincipal()) + "'s ");
        }
        name.append(((Name)s).getNameIndex(0));
        for (int i=1; i<((Name)s).size();i++) {
          name.append("'s " + ((Name)s).getNameIndex(i));
        }
      }
      else
        name.append(s.toShortString());
    }
    else if (o instanceof SDSIPublicKey) {
      name.append("[Key] ");
      name.append(getNameFor((SDSIPublicKey)o));
    }
    else {
      name.append(o.toShortString());
    }
    return name.toString();
  }
  public String getNameFor(SDSIPrincipal p) {
    SDSIPublicKey principal;
    if (p instanceof Hash) {
      principal = (SDSIPublicKey)get((Hash)p,KEY);
      if (principal == null)
        return "Unknown";
    }
    else 
      principal = (SDSIPublicKey)p;
    String s;
    if ((s = (String)keyNames.get(principal)) == null)
      return p.toShortString();
    else
      return s;
  }
  public void updateKeyNames(Def d) {
    //for now only add names for simple keys i define
    if (sameKeys(owner, d.getIssuer())) {
      if (d.getSubject() instanceof SDSIPublicKey) {
        keyNames.put(d.getSubject(), d.getName().getNameIndex(0));
      }
      else if (d.getSubject() instanceof Hash) {
        SDSIPublicKey k;
        if ((k = (SDSIPublicKey)allKey.get(d.getSubject())) != null) {
          keyNames.put(k,d.getName().getNameIndex(0));
        }
      }
    }
  }
  public void removeFromKeyNames(Def d) {
    if (sameKeys(owner, d.getIssuer())) {
      if (d.getSubject() instanceof SDSIPublicKey) {
        keyNames.remove(d.getSubject());
      }
      else if (d.getSubject() instanceof Hash) {
        SDSIPublicKey k;
        if ((k = (SDSIPublicKey)allKey.get(d.getSubject())) != null) {
          keyNames.remove(k);
        }
      }
    }
  }
  public Enumeration getDisplayedNames() {
    return displayedObjects.keys();
  }
  
  public SDSIObject select(String s) {
    Object o = displayedObjects.get(s);
    if (o == null)
      return null;
    else if (o instanceof CertSig)
      return ((CertSig)o).cert;
    else
      return (SDSIObject) o;
  }
  
  public boolean sameKeys(SDSIPublicKey k, SDSIPrincipal p) {
    return k.sameKeyAs(p);
  }
  
  private static Hash[] getAllHashes(SDSIObject obj) {
    Hash[] hashTypes = new Hash[definedHashTypes];
    try {  
      hashTypes[0] = new Hash("md5",obj);
      hashTypes[1] = new Hash("sha1",obj);
    }
    catch (NoSuchAlgorithmException e) {
    }
    return hashTypes;
  }
  
  public static final int KEY=0;
  public static final int AUTH=1;
  public static final int DEF=2;
  public static final int OTHER=3;
  public static final int UNKNOWN=4;
}