package sdsi.sexp;

import java.io.*;
import java.security.*;

class SimpleString implements Serializable{  
  byte length_bytes[];
  byte data[];
  
  /** Creates a new Simple string containing data**/
  SimpleString(byte[] data) {
    try {
      this.length_bytes = Integer.toString(data.length).getBytes("8859_1");
      this.data = data;
    }
    catch (UnsupportedEncodingException e){
      System.out.println("This shouldn't happen: Unsupported Encoding");}
  }
  /** Creates a new Simple string containing data**/
  SimpleString(String data) {
    try {
      this.length_bytes = Integer.toString(data.length()).getBytes("8859_1");
      this.data = data.getBytes("8859_1");
    }
    catch (UnsupportedEncodingException e){
      System.out.println("This shouldn't happen: Unsupported Encoding");}
  }
  /**Given an input stream, this parses off the first simple string S-expression
      from the input stream.**/
  public SimpleString(PushbackInputStream p) throws SexpException {
    int cur_byte;
    /* Do we want to watch the location of the reader when we have an exception
    */
    try {
      if ((cur_byte = p.read()) == '0') {   //Here we check for leading zero
        if ((cur_byte = p.read()) == ':') {  //Only allowed if its the entire number
          length_bytes = new byte[1];
          length_bytes[0] = (byte) '0';
          data = new byte[0];
        }
        else
          throw new SexpException("Unneccesary leading 0");
      }
      else {
        int length;
        if (Sexp.isDecimalDigit(cur_byte)) {  //there is a length
          p.unread(cur_byte);
          byte temp_length[]= new byte[20]; //Hardwire in limit of twenty digits in length(more than an int)
          int len_len = 0;
          while (Sexp.isDecimalDigit(cur_byte = p.read())) {
            temp_length[len_len++] = (byte) cur_byte;
          }
          length_bytes = new byte[len_len];
          for (int j=0;j<len_len;j++)
          length_bytes[j] = temp_length[j];
          length = Integer.parseInt(new String(length_bytes,"8859_1"));
          data = new byte[length];
          if (cur_byte == ':') {
            if (p.read(data,0,length) != length)
              throw new SexpException("Wrong time to end file, data too short");
          }
          else if (cur_byte == '#') {  //we have a length use it
            readHexDigits(length,p);  
          }
          else if (cur_byte == '|') {
            readBase64Digits(length,p);
          }
          else if (cur_byte == '\"') {
            readQuotedString(length,p);
          }
          else
            throw new SexpException("Token can not begin with decimal digit");
        }  //there is no length before the string
        else if (cur_byte == '#') {
          readHexDigits(p);
        }
        else if (cur_byte == '|') {
          readBase64Digits(p);
        }
        else if (cur_byte == '\"') {
          readQuotedString(p);
        }
        else {
          int bCount=0;
          data = new byte[16];
          byte[] oldData = data;
          if (Sexp.isTokenChar(cur_byte))
            data[bCount++]=(byte)cur_byte;
          else
            throw new SexpException("Unknown character attempted to be read as token");
          while(Sexp.isTokenChar(cur_byte = p.read())){
            data[bCount++]=(byte)cur_byte;
            if (bCount == data.length) {
              data = new byte[data.length *2];
              System.arraycopy(oldData,0,data,0,oldData.length);
              oldData = data;
            }
          }
          data = new byte[bCount];
          System.arraycopy(oldData,0,data,0,bCount);
          length_bytes = Integer.toString(data.length,10).getBytes("8859_1");
          p.unread(cur_byte);
        }
      }
    }
    catch (IOException e) {
      System.err.println("There has been an input/output error in SexpSimpString");
    }
  }
  private void readBase64Digits(PushbackInputStream p) throws SexpException{
    try {
      int cur_byte=0;
      byte[] bytes = new byte[16];
      byte[] oldBytes = bytes;
      int bCount=0;
      int numEquals = 0;
      while (true) {
        cur_byte = p.read();
        if (cur_byte == -1)
          throw new SexpException("Error in Base 64 rep, not long enough");
        else if (cur_byte == '|') {
          int numDone;
          if ((numDone =(bCount % 4)) == 3) {
            numEquals++;
            bytes[bCount++] = (byte) '=';
          }
          else if (numDone ==2 ) {
            numEquals+=2;
            bytes[bCount++] = (byte) '=';
            bytes[bCount++] = (byte) '=';
          }
          else if (numDone == 0)
            ;
          else if (numDone == 1)
            throw new SexpException("Error in Base 64 rep, not long enough");
          break;
        }
        else if (cur_byte == '=') {
          int numDone;
          if ((numDone =(bCount % 4)) == 3 || numDone == 2) {
            numEquals++;
            bytes[bCount] = (byte) '=';
            bCount++;
            if (bCount == bytes.length) {
              bytes = new byte[bytes.length + 1];  
              System.arraycopy(oldBytes,0,bytes,0,oldBytes.length);
              oldBytes = bytes;
            }
          }
          else
            throw new SexpException("Error in Base 64 rep, too many equals");
        }
        else if (Sexp.isBase64Digit(cur_byte)) {
          if (numEquals > 0)
            throw new SexpException("Error in Base64 rep, = in the middle");
          bytes[bCount]=(byte) cur_byte;
          bCount++;
          if (bCount == bytes.length) {
            bytes = new byte[bytes.length *2];
            System.arraycopy(oldBytes,0,bytes,0,oldBytes.length);
            oldBytes = bytes;
          }
        }
        else if (Sexp.isWhiteSpace(cur_byte))
          continue;
        else
          throw new SexpException("Error, non-Base64 character in Base64 rep");
      }
      if (cur_byte != '|')
        throw new SexpException("Error, Base64 representation ended mistakenly"); // remove
      bytes = new byte[bCount];  //bCount should be mult of four here
      System.arraycopy(oldBytes,0,bytes,0,bCount);
      data = new byte[3*(bCount/4)-numEquals];
      byte[] threeBytes;
      for (int i=0; i< bCount/4;i++) {
        threeBytes = Sexp.base64ToOctet(bytes[4*i],bytes[4*i+1],bytes[4*i+2],bytes[4*i+3]);
        for (int j=0; j < threeBytes.length; j++)
          data[3*i+j] = threeBytes[j];
      }
      length_bytes = Integer.toString(data.length,10).getBytes("8859_1");
    }
    catch (IOException e) {
      System.err.println("There has been an input/output error reading a base64 rep in SexpSimpString");
    } 
  }
  private void readQuotedString(PushbackInputStream p) throws SexpException{
    try {
      int cur_byte;
      int bCount = 0;
      data = new byte[16];
      byte[] oldData = data;
      while ((cur_byte = p.read()) != '\"') {
        if (cur_byte == -1)
          throw new SexpException("Error in quoted string rep, no ending quote");
        else if (cur_byte == '\\') {
          bCount=processEscapeSequence(bCount,p); //manipulate data
        } 
        else {
          data[bCount++] = (byte) cur_byte;
        }
        if (bCount == data.length) {
          data = new byte[data.length *2];
          System.arraycopy(oldData,0,data,0,oldData.length);
          oldData = data;
        }
      }
      data = new byte[bCount];
      System.arraycopy(oldData,0,data,0,bCount);
      length_bytes = Integer.toString(data.length,10).getBytes("8859_1");
    }
    catch (IOException e) {
      System.err.println("There has been an input/output error reading a quoted string rep in SexpSimpString");
    }
  }
  private void readHexDigits(PushbackInputStream p) throws SexpException{
    try {
      int cur_byte;
      byte[] bytes = new byte[16];
      byte[] oldbytes = bytes;
      int bCount = 0;
      while ((cur_byte = p.read()) != '#') {
        if (cur_byte == -1)
          throw new SexpException("Error in hexadecimal rep");
        else if (Sexp.isHexDigit(cur_byte)) {
          bytes[bCount]=(byte) cur_byte;
          bCount++;
          if (bCount == bytes.length) {
            bytes = new byte[bytes.length *2];
            System.arraycopy(oldbytes,0,bytes,0,oldbytes.length);
            oldbytes = bytes;
          }
        }
        else if (Sexp.isWhiteSpace(cur_byte)) 
          continue;
        else
          throw new SexpException("Error, non-Hexadecimal character in hex rep");
      }
      bytes = new byte[bCount];
      System.arraycopy(oldbytes,0,bytes,0,bCount);
      String number = new String(bytes, "8859_1");
      if ((number.length() % 2) != 0)
        throw new SexpException("Odd number of hex digits in representation.");
      data = new byte[number.length() / 2];
      for (int i=0;i<data.length;i++)
        data[i] = Integer.valueOf(number.substring(2*i,2*i+2),16).byteValue();
      length_bytes = Integer.toString(data.length,10).getBytes("8859_1");
    }
    catch (IOException e) {
      System.err.println("There has been an input/output error reading a hex rep in SexpSimpString");
    }
  }
  private void readBase64Digits(int length, PushbackInputStream p) throws SexpException{
    try {
      int cur_byte=0;
      int num64chars = 4*(int)Math.ceil((double)length/3.0);
      byte[] bytes = new byte[num64chars];
      int bCount=0;
      boolean seenEquals = false;
      while (bCount<num64chars) {
        cur_byte = p.read();
        if (cur_byte == -1)
          throw new SexpException("Error in Base 64 rep, not long enough");
        else if (cur_byte == '|') {
          int numLeft;
          if ((numLeft = num64chars - bCount) == 1 || numLeft == 2) {
            bytes[num64chars - numLeft] = bytes[num64chars - 1] = (byte) '=';
            bCount = num64chars;
            break;
          }
          else
            throw new SexpException("Error in Base 64 rep, not long enough");
        }
        else if (cur_byte == '=') {
          int numLeft;
          if ((numLeft = num64chars - bCount) == 1 || numLeft == 2) {
            seenEquals = true;
            bytes[bCount] = (byte) '=';
            bCount++;
          }
          else
            throw new SexpException("Error in Base 64 rep, not long enough");
        }
        else if (Sexp.isBase64Digit(cur_byte)) {
          if (seenEquals)
            throw new SexpException("Error in Base64 rep, = in the middle");
          bytes[bCount]=(byte) cur_byte;
          bCount++;
        }
        else if (Sexp.isWhiteSpace(cur_byte))
          continue;
        else
          throw new SexpException("Error, non-Base64 character in Base64 rep");
      }
      if (cur_byte != '|') {
        while (Sexp.isWhiteSpace(cur_byte = p.read()))
          ;
        if (cur_byte != '|')
          throw new SexpException("Error, Base64 representation didn't end after length");
      }
      byte[] threeBytes;
      for (int i=0; i< num64chars/4;i++) {
        threeBytes = Sexp.base64ToOctet(bytes[4*i],bytes[4*i+1],bytes[4*i+2],bytes[4*i+3]);
        for (int j=0; j < threeBytes.length; j++)
          data[3*i+j] = threeBytes[j];
      }
    }
    catch (IOException e) {
      System.err.println("There has been an input/output error reading a base64 rep in SexpSimpString");
    }           
  }
  private final int processEscapeSequence(int bCount, PushbackInputStream p) {
    try {
      int cur_byte = p.read();
      String number;
      switch(cur_byte) {
        case 'b':
          data[bCount++]='\b';
          break;
        case 't':
          data[bCount++]='\t';
          break;
        case 'v':
          data[bCount++]='\t'; //no vertical tab!!
          break;
        case 'f':
          data[bCount++]='\f';
          break;  
        case 'n':
          data[bCount++]='\n';
          break;
        case 'r':
          data[bCount++]='\r';
          break;
        case '\'':
          data[bCount++]='\'';
          break;
        case '\"':
          data[bCount++]='\"';
          break;
        case '\\':
          data[bCount++]='\\';
          break;  
        case 'x':
          byte[] hexRep = new byte[2];
          hexRep[0] = (byte) p.read(); //probably not exactly right white space might work or something
          hexRep[1] = (byte) p.read();
          number = new String(hexRep, "8859_1");
          data[bCount++] = Integer.valueOf(number,16).byteValue();//watch number format exception, I think its IO
          break;
        case '\n':
          if ((cur_byte = p.read()) == '\r')
            ;
          else
            p.unread(cur_byte);
          break;
        case '\r':
          if ((cur_byte = p.read()) == '\n')
            ;
          else
            p.unread(cur_byte);
          break;
        default:
          byte[] octRep = new byte[3];
          octRep[0] = (byte) cur_byte;
          octRep[1] = (byte) p.read(); //probably not exactly right white space might work or something
          octRep[2] = (byte) p.read();
          number = new String(octRep, "8859_1");
          data[bCount++] = Integer.valueOf(number,8).byteValue();//watch number format exception, I think its IO
          break;
      }
    }
    catch (IOException e) {
      System.err.println("There has been an input/output error reading a quoted string rep in SexpSimpString");
    }
    return bCount;
  }
  private void readQuotedString(int length, PushbackInputStream p) throws SexpException{
    try {
      int cur_byte;
      int bCount = 0;
      while (bCount < length) {
        cur_byte=p.read();
        if (cur_byte == -1 || cur_byte == '\"')
          throw new SexpException("Error in quoted string rep, not long enough");
        else if (cur_byte == '\\') {
          bCount=processEscapeSequence(bCount,p); //manipulate data
        } 
        else {
          data[bCount++] = (byte) cur_byte;
        }
      }
      if ((cur_byte = p.read()) != '\"')
        throw new SexpException("Error in quoted string rep, too long");
    }
    catch (IOException e) {
      System.err.println("There has been an input/output error reading a quoted string rep in SexpSimpString");
    }
  }
  private void readHexDigits(int length, PushbackInputStream p) throws SexpException{
    try {
      int cur_byte;
      byte[] bytes = new byte[length*2];
      int bCount = 0;
      while (bCount<length*2) {
        cur_byte = p.read();
        if (cur_byte == -1 || cur_byte == '#')
          throw new SexpException("Error in hexadecimal rep, not long enough");
        else if (Sexp.isHexDigit(cur_byte)) {
          bytes[bCount]=(byte) cur_byte;
          bCount++;
        }
        else if (Sexp.isWhiteSpace(cur_byte))
          continue;
        else
          throw new SexpException("Error, non-Hexadecimal character in hex rep");
      }
      while (Sexp.isWhiteSpace(cur_byte = p.read()))
        ;
      if (cur_byte != '#')
        throw new SexpException("Error, hexadecimal representation didn't end after length");
      String number = new String(bytes, "8859_1");
      for (int i=0;i<length;i++)
        data[i] = Integer.valueOf(number.substring(2*i,2*i+2),16).byteValue();
    }
    catch (IOException e) {
      System.err.println("There has been an input/output error reading a hex rep in SexpSimpString");
    }
  }
  boolean canBeTokenString() {
    if (data.length==0)
      return false;
    if (Sexp.isDecimalDigit(data[0]))
      return false;
    for (int i=0;i<data.length;i++) {
      if (!Sexp.isTokenChar(data[i]))
        return false;
    }
    return true;
  }
  boolean canBeQuotedString() {
    for (int i=0;i<data.length;i++){
      if (!Sexp.isTokenChar(data[i]) && !(data[i] == ' '))
        return false;
    }
    return true;
  }
  StringBuffer toTokenString() {
    String s = null;
    try {
      s = new String(data,"8859_1"); 
    }
    catch (UnsupportedEncodingException e){
      //do something this can't happen
    }
    return new StringBuffer(s);
  }
  StringBuffer toQuotedString() {
    StringBuffer s = this.toTokenString();
    s.insert(0,'\"');
    s.append('\"');
    return s;
  }
  StringBuffer toHexString() {
    StringBuffer s = new StringBuffer();
    s.append("#");
    for (int i=0;i<data.length;i++) {
      s.append(Sexp.hexDigit[data[i]>>>4 & 0x0F]);
      s.append(Sexp.hexDigit[data[i] & 0x0F]);
    }
    s.append("#");
    return s;
  }
  StringBuffer toBase64String() {
    StringBuffer s= new StringBuffer();
    s.append("|");
    int carry = 0;
    for (int i=0;i<data.length;i++) {
      switch (i%3) {
      case 0:
        s.append(Sexp.base64Digit[data[i]>>>2 & 0x3F]);
        carry = data[i]<<4 & 0x30;
        break;
      case 1:
        s.append(Sexp.base64Digit[carry + (data[i]>>>4 & 0x0F)]);
        carry = data[i]<<2 & 0x3C;
        break;
      case 2:
        s.append(Sexp.base64Digit[carry + (data[i]>>>6 & 0x03)]);
        s.append(Sexp.base64Digit[data[i] & 0x3F]);
        carry = 0;
        break;
      }
    }
    if (data.length%3 == 1) {
      s.append(Sexp.base64Digit[carry & 0x3F]);
      s.append('=');
      s.append('=');
    }
    if (data.length%3 == 2) {
      s.append(Sexp.base64Digit[carry & 0x3F]);
      s.append('=');
    }
    s.append('|');
    return s;
  }      
  //this can be done better by only scanning input once
  //maybe somehow combine this with getReadableLen
  //first time either one is called store how it can
  //be represented  thats good!!
  StringBuffer toReadableStringBuffer() {
    if (this.canBeTokenString())
      return this.toTokenString();
    else if (this.canBeQuotedString())
      return this.toQuotedString();
    else if (this.data.length < 8)
      return this.toHexString();
    else
      return this.toBase64String();
  }
  int getReadableLen() {
    if (this.canBeTokenString())
      return data.length;
    else if (this.canBeQuotedString())
      return data.length + 2;
    else if (this.data.length < 8)
      return 2*data.length;
    else
      return 4*((data.length + 2)/3)+2;
  }
  public String toVerbatimString() {
    String s = null;
    try {
      s = new String(this.getCanonRep(),"8859_1");
    }
    catch (UnsupportedEncodingException e){
      //this can't happen
    }
    return s;
  }
  public void display() {
    System.out.print("Length is: " + data.length); 
    System.out.print("\tData is: ");
    for (int i= 0; i< data.length; i++)
      System.out.print((char) data[i]);
    System.out.print("\n");
  }
  public int getCanonLen() {
    return length_bytes.length+data.length+1;
  }
  public byte[] getCanonRep() {
    int len,len_len,i;
    len = this.getCanonLen();
    len_len = length_bytes.length;
    byte[] canon = new byte[len];
    System.arraycopy(length_bytes,0,canon,0,len_len);
    canon[len_len] = (byte) ':';
    System.arraycopy(data,0,canon,len_len+1,data.length);
    return canon;
  }
  public boolean equals(SimpleString other) {
    return MessageDigest.isEqual(this.data,other.data);
  }
  public int compareTo(SimpleString other) {
    return Sexp.toString(this.data).compareTo(Sexp.toString(other.data));
  }
  public boolean startsWith(SimpleString other) {
    return Sexp.toString(this.data).startsWith(Sexp.toString(other.data));
  }
}  
